/* Virtual disk driver for Syllable.
   Currently supports VirtualPC fixed-size VHD files.
   Use the 'mountvhd' util to mount and unmount virtual disks. 			*/
   
/* Written by Anthony Morphett  awmorp@gmail.com    July 2007     */

/* Known bugs & limitations:
	- you cannot unmount a virtual disk if you have previously tried unsuccessfully to mount one of the partitions as ext2
		(this happens if you don't specify a filesystem to the mount command, and ext2 probes the partition automatically).
		This is a bug in the ext2 driver which should be fixed in the next Syllable release (or in a weekly build).
	- a maximum of 16 partitions are supported. (I don't know if it's even possible to have more than 16 partitions on
		one disk, but if you do, only the first 16 will be available).
	- changes to the partition table of a virtual disk are not detected until the next mount.
*/

/* TODO:
	  * use system decode_partition_table function
	  * support dynamically expanding disks as well as fixed-size disks
	  * VMWare virtual disks	  
*/

#include <posix/errno.h>
#include <posix/unistd.h>  /* for SEEK_END */

#include <atheos/types.h>
#include <atheos/device.h>
#include <atheos/kernel.h>
#include <atheos/semaphore.h>
#include <atheos/spinlock.h>

#include <macros.h>

#include "virtualdisk.h"
#include "protocol.h"


sem_id g_hDiskLock;

VirtualDisk_s* g_psFirstVDisk = NULL;

DeviceOperations_s g_sVDiskOps = {
	vdisk_open,
	vdisk_close,
	vdisk_ioctl,
	vdisk_read,
	vdisk_write,
	vdisk_readv,
	vdisk_writev,
	NULL,
	NULL
};

/* Operations for the control node /dev/disk/virtual/control - only IOCTL */
DeviceOperations_s g_sVControlOps = {
	vdisk_control_open,
	vdisk_control_close,
	vdisk_control_ioctl,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

int g_nDeviceID;
int g_nControlDevNodeHandle;

/* Swap the byte order */
void byteswap( void* pWhere, int nSize )
{
	uint8 nTemp;
	int i = 0;
	for( ; i < (nSize+1) / 2; i++ )
	{
		nTemp = ((uint8*)pWhere)[i];
		((uint8*)pWhere)[i] = ((uint8*)pWhere)[nSize-1-i];
		((uint8*)pWhere)[nSize-1-i] = nTemp;
	}
}

int pread( int nHandle, void* pBuffer, size_t nLength, size_t nOffset )
{
	int nResult;
	lseek( nHandle, nOffset, SEEK_SET );
	nResult = read( nHandle, pBuffer, nLength );
	return( nResult );
}

int pwrite( int nHandle, const void* pBuffer, size_t nLength, size_t nOffset )
{
	int nResult;
	lseek( nHandle, nOffset, SEEK_SET );
	nResult = write( nHandle, pBuffer, nLength );
	return( nResult );
}

/* Read some data from the virtual disk */
int vdisk_pread( VirtualDisk_s* psVDisk, void* pBuffer, size_t nLength, size_t nOffset )
{
	/* TODO: This is trivial for fixed-size disks, but lookup code for dynamically-expanding disks would go here */
	int nResult;
	nResult = pread( psVDisk->m_nHandle, pBuffer, nLength, nOffset );
	return( nResult );
}

/* Write data to the virtual disk at the specified location */
int vdisk_pwrite( VirtualDisk_s* psVDisk, const void* pBuffer, size_t nLength, size_t nOffset )
{
	/* TODO: dynamic disks */
	int nResult;
	nResult = pwrite( psVDisk->m_nHandle, pBuffer, nLength, nOffset );
	return( nResult );
}

uint32 compute_checksum( VHDFooter_s* psFooter )
{
	uint32 nChecksum = 0;
	uint32 nOldChecksum = psFooter->m_nChecksum;
	psFooter->m_nChecksum = 0;
	
	int i = 0;
	for( ; i < sizeof( VHDFooter_s ); i++ ) {
		nChecksum += ((uint8*)psFooter)[i];
	}
	nChecksum = ~nChecksum;
	psFooter->m_nChecksum = nOldChecksum;
	return( nChecksum );
}

/* Callback funtion for decode_disk_partitions() */
size_t vdisk_read_callback( void* pCookie, off_t nOffset, void* pBuffer, size_t nSize )
{
	return( vdisk_pread( (VirtualDisk_s*)pCookie, pBuffer, nSize, nOffset ) );
}

#if 0
/* TODO: use driver API decode_partition_table() function (in kernel/vfs/dev.c) */
/* Read the partition table and create device nodes for each partition. Return the number of partitions. */
int read_vhd_partition_table( VirtualDisk_s* psVDisk, size_t nMBRLocation, size_t nDiskSize, int nPartitionsReadSoFar )
{
	int nResult;
	MBR_s sMBR;
	
	nResult = vdisk_pread( psVDisk, &sMBR, sizeof( MBR_s ), nMBRLocation );
	if( nResult != sizeof( MBR_s ) ) {
		printk("VDisk: could not read MBR & partition table from disk %s.\n", psVDisk->m_pzFileName );
		return( 0 );
	}
#if 0
	printk( "sizeof( MBR_s ) = %i   offset of m_nMBRId = %i   sizeof( MBRPartitionEntry ) = %i  sizeof( MBRId ) = %i\n", sizeof( MBR_s ), ((int)&sMBR.m_nMBRId) - ((int)&sMBR), sizeof( MBRPartitionEntry_s ), sizeof( sMBR.m_nMBRId ) );
	printk( "MBR Dump:\n" );
	char zTmp[65];
	zTmp[65] = 0;
	int j = 0;
	int k = 0;
	for( ; k < 16; k++ )
	{
		for( j = 0; j < 32; j++ ) { sprintf( &zTmp[2*j], "%02x", (int)(((uint8*)&sMBR)[j+32*k]) ); }
		zTmp[65] = 0;
		printk( "%04x: %s\n", k*32, zTmp );
	}
#endif
	if( sMBR.m_nMBRId != 0xaa55 ) {
		printk( "VDisk: MBR signature 0xaa55 is missing from sector %i of disk %s (it has 0x%04x). Perhaps the MBR is corrupt or missing. Attempting to decode partition table nevertheless.\n", nMBRLocation / 512, psVDisk->m_pzFileName, sMBR.m_nMBRId );
	}
	int i = 0;
	for( ; i < 4; i++ ) {
		int nStartCyl, nStartSec, nEndCyl, nEndSec;
		nStartSec = sMBR.m_aPartitionEntry[i].m_nStartCylSec & 0x3f;
		nStartCyl = (sMBR.m_aPartitionEntry[i].m_nStartCylSec >> 8) + ((sMBR.m_aPartitionEntry[i].m_nEndCylSec & 0xc0)<<10);
		nEndSec = sMBR.m_aPartitionEntry[i].m_nEndCylSec & 0x3f;
		nEndCyl = (sMBR.m_aPartitionEntry[i].m_nEndCylSec >> 8) + ((sMBR.m_aPartitionEntry[i].m_nEndCylSec & 0xc0)<<10);
		if( sMBR.m_aPartitionEntry[i].m_nPartitionType != 0 ) {
			printk( "VDisk: partition %i: status=%i  start head/cyl/sec=%i/%i/%i  type=%02x end head/cyl/sec=%i/%i/%i start sector=%i  length=%i\n",
			  i, sMBR.m_aPartitionEntry[i].m_nStatus, sMBR.m_aPartitionEntry[i].m_nStartHead, nStartCyl, nStartSec, sMBR.m_aPartitionEntry[i].m_nPartitionType, sMBR.m_aPartitionEntry[i].m_nEndHead, nEndCyl, nEndSec, sMBR.m_aPartitionEntry[i].m_nLBAStart, sMBR.m_aPartitionEntry[i].m_nLBALength );
		}

		/* TODO: sanity checking - check for overlaps, etc */
		if( sMBR.m_aPartitionEntry[i].m_nPartitionType == 0x05 || sMBR.m_aPartitionEntry[i].m_nPartitionType == 0x0f ) {  /* Extended partition */
			printk( "VDisk: Found extended partition at sector %i\n", (nMBRLocation/512) + sMBR.m_aPartitionEntry[i].m_nLBAStart );
				nPartitionsReadSoFar = read_vhd_partition_table( psVDisk, nMBRLocation + sMBR.m_aPartitionEntry[i].m_nLBAStart*512, sMBR.m_aPartitionEntry[i].m_nLBALength * 512, nPartitionsReadSoFar );
		} else if( sMBR.m_aPartitionEntry[i].m_nPartitionType != 0 ) {  /* Data partition */
			if( nPartitionsReadSoFar >= 17 ) { printk( "VDisk: Currently a maximum of 16 partitions are supported. Partition %i will not be available.\n", nPartitionsReadSoFar ); }
			else { /* Add it to the list and create a device node */
				psVDisk->m_asPartitions[nPartitionsReadSoFar].m_psDisk = psVDisk;
				psVDisk->m_asPartitions[nPartitionsReadSoFar].m_nIndex = nPartitionsReadSoFar;
				psVDisk->m_asPartitions[nPartitionsReadSoFar].m_nStart = nMBRLocation + sMBR.m_aPartitionEntry[i].m_nLBAStart * 512;
				psVDisk->m_asPartitions[nPartitionsReadSoFar].m_nSize = sMBR.m_aPartitionEntry[i].m_nLBALength * 512;
				psVDisk->m_asPartitions[nPartitionsReadSoFar].m_nStatus = sMBR.m_aPartitionEntry[i].m_nStatus;
				psVDisk->m_asPartitions[nPartitionsReadSoFar].m_nPartitionType = sMBR.m_aPartitionEntry[i].m_nPartitionType;
				atomic_set( &psVDisk->m_asPartitions[nPartitionsReadSoFar].m_nOpenCount,  0);  /* TODO: should use atomic_set() or something */
				psVDisk->m_nNumPartitions++;
				
				/* create the device node */
				char zDevNode[255];
				sprintf( zDevNode, "%s/%i", psVDisk->m_pzDevDir, nPartitionsReadSoFar - 1 ); /* TODO: use snprintf */
				nResult = create_device_node( g_nDeviceID, -1, zDevNode, &g_sVDiskOps, &(psVDisk->m_asPartitions[nPartitionsReadSoFar]) );
				printk( "created node. handle=%i cookie=%x disk=%i\n", nResult, (int)&(psVDisk->m_asPartitions[nPartitionsReadSoFar]), (int)(psVDisk) );
				if( nResult <= 0 ) {
					printk( "VDisk: could not create device node %s!\n" );
					psVDisk->m_asPartitions[nPartitionsReadSoFar].m_nDevNodeHandle = 0;
				}
				else { psVDisk->m_asPartitions[nPartitionsReadSoFar].m_nDevNodeHandle = nResult; }
			}
			nPartitionsReadSoFar++;
		}
	}
	return( nPartitionsReadSoFar );
}
#else
/* Use decode_partition_table() to verify and read the disk's partitions */
int get_vhd_partitions( VirtualDisk_s* psVDisk )
{
//	printk( "get_vhd_partitions(_)\n" );
	int nResult;
	Partition_s asPartitionData[16];
	device_geometry sGeom;
	sGeom.sector_count = psVDisk->m_nDiskSize / 512;
	sGeom.cylinder_count = 1;
	sGeom.sectors_per_track = 1;
	sGeom.head_count = 1;
	sGeom.bytes_per_sector = 512;
	sGeom.read_only = psVDisk->m_bReadOnly;
	sGeom.removable = false;
	
	nResult = decode_disk_partitions( &sGeom, asPartitionData, 16, psVDisk, vdisk_read_callback );
	if( nResult < 0 ) {
		printk( "VDisk: error reading partitions. Perhaps the partition table is corrupt or the disk has not been partitioned.\n" );
		return( nResult );
	}
	int i = 0;
	int nCount = 0;
	char zTmp[256];
	/* Copy the info into the VirtualPartition_s array */
	for( ; i < nResult; i++ ) {
		/* if the type is 0, it is a hole in the partition table */
		psVDisk->m_asPartitions[i].m_psDisk = psVDisk;
		psVDisk->m_asPartitions[i].m_nIndex = i+1;   /* the 0'th element is 'raw' - the whole disk */
		if( asPartitionData[i].p_nType != 0 && asPartitionData[i].p_nSize != 0 ) {
			psVDisk->m_asPartitions[i].m_nPartitionType = asPartitionData[i].p_nType;
			psVDisk->m_asPartitions[i].m_nStatus = asPartitionData[i].p_nStatus;
			psVDisk->m_asPartitions[i].m_nStart = asPartitionData[i].p_nStart;
			psVDisk->m_asPartitions[i].m_nSize = asPartitionData[i].p_nSize;
			atomic_set( &psVDisk->m_asPartitions[i].m_nOpenCount, 0 );
			sprintf( zTmp, "%s/%i", psVDisk->m_pzDevDir, i );
			psVDisk->m_asPartitions[i].m_nDevNodeHandle = create_device_node( g_nDeviceID, -1, zTmp, &g_sVDiskOps, &(psVDisk->m_asPartitions[i]) );
			if( psVDisk->m_asPartitions[i].m_nDevNodeHandle < 0 ) {
				nCount++;
				printk( "VDisk: could not create device node %s!\n", zTmp );
				psVDisk->m_asPartitions[i].m_nPartitionType = 0;
			}
		} else {
			psVDisk->m_asPartitions[i].m_nPartitionType = 0;
			psVDisk->m_asPartitions[i].m_nDevNodeHandle = -1;
		}
	}
	for( ; i < 16 ; i++ ) {
		psVDisk->m_asPartitions[i].m_psDisk = psVDisk;
		psVDisk->m_asPartitions[i].m_nIndex = i+1;
		psVDisk->m_asPartitions[i].m_nPartitionType = 0;
		psVDisk->m_asPartitions[i].m_nDevNodeHandle = 0;
	}
	psVDisk->m_nNumPartitions = nCount;
	return( nCount );
}
#endif

/* Verify that the file is a valid VHD file */
bool verify_vhd_footer( VirtualDisk_s* psVDisk )
{
//	printk( "VDisk: verify_vhd( %s )\n", psVDisk->m_pzFileName );
	int nResult;
	
	VHDFooter_s* psFooter = &psVDisk->m_sFooter;

	/* It seems that doing lseek( _, -sizeof( VHDFooter_s ), SEEK_END ) doesn't work!? It causes the read() call to fail. */
//	nResult = lseek( psVDisk->m_nHandle, 157286400/*sizeof( VHDFooter_s )*/, SEEK_SET );
	nResult = lseek( psVDisk->m_nHandle, 0, SEEK_END );
	if( nResult < 0 ) {printk( "VDisk: verify_vhd seek failed!? %i\n", nResult );}
	nResult = pread( psVDisk->m_nHandle, psFooter, sizeof( VHDFooter_s ), nResult - sizeof( VHDFooter_s ) );
	if( nResult != sizeof( VHDFooter_s ) ) {
		printk( "VDisk: could not read footer for %s, perhaps it is not a valid VHD file. result is %i\n", psVDisk->m_pzFileName, nResult );
		return( false );
	}
	byteswap( &psFooter->m_nFeatures, 4 );
	byteswap( &psFooter->m_nFileFormatVersion, 4 );
	byteswap( &psFooter->m_nDataOffset, 8 );
	byteswap( &psFooter->m_nTimeStamp, 4 );
	byteswap( &psFooter->m_nCreatorAppVersion, 4 );
	byteswap( &psFooter->m_nOriginalSize, 8 );
	byteswap( &psFooter->m_nCurrentSize, 8 );
	byteswap( &psFooter->m_sDiskGeometry.m_nCylinders, 2 );
	byteswap( &psFooter->m_nDiskType, 4 );
	byteswap( &psFooter->m_nChecksum, 4 );
#if 0
	printk( "VDisk: read the footer from %s.\n", psVDisk->m_pzFileName );
	printk( " cookie = '%c%c%c%c%c%c%c%c' features = %08x file format version = %08x data offset = %li time stamp = %i\n", psFooter->m_zCookie[0], psFooter->m_zCookie[1], psFooter->m_zCookie[2], psFooter->m_zCookie[3], psFooter->m_zCookie[4], psFooter->m_zCookie[5], psFooter->m_zCookie[6], psFooter->m_zCookie[7], psFooter->m_nFeatures, psFooter->m_nFileFormatVersion, psFooter->m_nDataOffset, psFooter->m_nTimeStamp );
	printk( " creator app = '%c%c%c%c' creator app version = %08x creator OS host = '%c%c%c%c'\n", psFooter->m_zCreatorApp[0], psFooter->m_zCreatorApp[1], psFooter->m_zCreatorApp[2], psFooter->m_zCreatorApp[3], psFooter->m_nCreatorAppVersion, psFooter->m_zCreatorOSHost[0], psFooter->m_zCreatorOSHost[1], psFooter->m_zCreatorOSHost[2], psFooter->m_zCreatorOSHost[3] );
	printk( " original size = %li current size = %li disk geometry: %i cylinders, %i heads, %i sectors per track\n", psFooter->m_nOriginalSize, psFooter->m_nCurrentSize, psFooter->m_sDiskGeometry.m_nCylinders, psFooter->m_sDiskGeometry.m_nHeads, psFooter->m_sDiskGeometry.m_nSectorsPerTrack );
	char zUUID[40];
	int i = 0;
	for( ; i < 16; i++ ) {
		sprintf( &zUUID[2*i], "%08x", psFooter->m_aUUID[i] );
	}
	zUUID[32] = 0;
	printk( " disk type = %08x checksum = %08x UUID = %s saved state = %i\n", psFooter->m_nDiskType, psFooter->m_nChecksum, zUUID, psFooter->m_nSavedState );
#endif
	
	/* check file format */
	if( (psFooter->m_nFileFormatVersion >> 16) != 1 ) {
		printk( "VDisk: Warning: file %s has unsupported file format version %i - results may be unexpected!\n", psVDisk->m_pzFileName, (psFooter->m_nFileFormatVersion >> 16) );
		/* TODO: make sure we mount read-only */
	}
	
	/* verify the checksum */
	int nChecksum = compute_checksum( psFooter );
//	printk( "VDisk: computed checksum %08x\n", nChecksum );
	if( nChecksum != psFooter->m_nChecksum ) {
		printk( "VDisk: Warning: checksum failed! The virtual disk is possibly corrupt.\n" );
	}
	
	if( psFooter->m_nDiskType != 2  ) {
		printk( "VDisk: Error: file %s is an unsupported disk type. Currently only fixed-size disks are supported.\n", psVDisk->m_pzFileName );
		/* TODO: free footer */
		return( false );
	}
	if( psFooter->m_nSavedState != 0 && !psVDisk->m_bReadOnly ) {
		printk( "VDisk: Warning: file %s is attached to a virtual machine in saved state. Modifying the virtual disk may cause unexpected results if the virtual machine is resumed. It is recommended to mount the disk read-only.\n", psVDisk->m_pzFileName );
	}
	
//	printk( "VDisk: file %s is a valid virtual disk.\n", psVDisk->m_pzFileName );
	return( true );
}

status_t mount_vhd( char* pzFile, char* pzMountDir, bool bReadOnly )
{
	printk( "VDisk: mounting %s\n", pzFile );
	int nResult;
		
	LOCK( g_hDiskLock );
	VirtualDisk_s* psVDisk = kmalloc( sizeof( VirtualDisk_s ), MEMF_KERNEL );
	if( !psVDisk ) { printk( "VDisk: insufficient memory to mount the disk!\n" ); UNLOCK( g_hDiskLock ); return( -ENOMEM ); }

	psVDisk->m_hLock = create_semaphore( "vdisk_semaphore", 1, 0 );
	if( psVDisk->m_hLock < 0 ) {
		printk( "VDisk: could not create semaphore for disk %s!\n", pzFile );
		nResult = psVDisk->m_hLock;
		kfree( psVDisk );
		UNLOCK( g_hDiskLock );
		return( nResult );
	}
	nResult = open( pzFile, (bReadOnly ? O_RDONLY : 0) );
	if( nResult <= 0 ) {
		printk( "VDisk: could not open file %s! result is %i\n", pzFile, nResult );
		kfree( psVDisk );
		UNLOCK( g_hDiskLock );
		return( nResult );
	}
	psVDisk->m_nHandle = nResult;
	psVDisk->m_bReadOnly = bReadOnly;
	psVDisk->m_pzFileName = kmalloc( strlen( pzFile ), MEMF_KERNEL );
	if( psVDisk->m_pzFileName == NULL ) {
		printk( "VDisk: insufficient memory to mount the disk!\n" );
		close( psVDisk->m_nHandle );
		kfree( psVDisk );
		UNLOCK( g_hDiskLock );
		return( -ENOMEM );
	}
	strcpy( psVDisk->m_pzFileName, pzFile );
	if( !verify_vhd_footer( psVDisk ) ) {
		printk( "VDisk: %s is not a valid VHD file!\n", pzFile );
		close( psVDisk->m_nHandle );
		kfree( psVDisk->m_pzFileName );
		kfree( psVDisk );
		UNLOCK( g_hDiskLock );
		return( -EINVAL );
	}
	
	psVDisk->m_nDiskSize = psVDisk->m_sFooter.m_nCurrentSize;
	
	char* pzTmp = pzMountDir;
	if( pzTmp == NULL || pzTmp[0] == 0 ) {
		pzTmp = pzFile;
	}
	pzMountDir = strrchr( pzTmp, '/' );
	if( !pzMountDir ) { pzMountDir = pzTmp; }
	else { pzMountDir++; }  /* skip the '/' */
	
	psVDisk->m_pzDevDir = kmalloc( strlen( pzMountDir ) + strlen( "disk/virtual/" ) + 1, MEMF_KERNEL );
	if( !psVDisk->m_pzDevDir ) {
		printk( "VDisk: out of memory!\n" );
		close( psVDisk->m_nHandle );
		kfree( psVDisk->m_pzFileName );
		kfree( psVDisk );
		UNLOCK( g_hDiskLock );
		return( -ENOMEM );
	}
	sprintf( psVDisk->m_pzDevDir, "disk/virtual/%s", pzMountDir );
	pzMountDir = psVDisk->m_pzDevDir;

	/* Create the raw partition */
	psVDisk->m_asPartitions[0].m_psDisk = psVDisk;
	psVDisk->m_asPartitions[0].m_nIndex = 0;
	psVDisk->m_asPartitions[0].m_nStart = 0;
	psVDisk->m_asPartitions[0].m_nSize = psVDisk->m_nDiskSize;
	
	char zDevNode[256];
	sprintf( zDevNode, "%s/raw", pzMountDir );  /* TODO: use snprintf */
    nResult = create_device_node( g_nDeviceID, -1, zDevNode, &g_sVDiskOps, &psVDisk->m_asPartitions[0] );
	if( nResult <= 0 ) {
		printk( "VDisk: failed to create device node %s! result is %i\n", zDevNode, nResult );
		close( psVDisk->m_nHandle );
		kfree( psVDisk->m_pzFileName );
		kfree( psVDisk->m_pzDevDir );
		kfree( psVDisk );
		UNLOCK( g_hDiskLock );
		return( -EINVAL );
	}
	psVDisk->m_nNumPartitions = 1;
	psVDisk->m_asPartitions[0].m_nDevNodeHandle = nResult;
	
	LOCK( psVDisk->m_hLock );
#if 0
	read_vhd_partition_table( psVDisk, 0, psVDisk->m_nDiskSize, 1 );  /* Read the partition table and create device nodes for each partition, recursing into extended partitions */
#else
	get_vhd_partitions( psVDisk );   /* Read & verify the partition table, and create device nodes for all valid partitions */
#endif

	psVDisk->m_psNextVDisk = g_psFirstVDisk;
	g_psFirstVDisk = psVDisk;
	printk( "VDisk: successfully mounted %s at %s\n", pzFile, pzMountDir );
	UNLOCK( psVDisk->m_hLock );
	UNLOCK( g_hDiskLock );
	return( 0 );
}

status_t unmount_vhd( VirtualDisk_s* psVDisk )
{
	printk( "VDisk: unmounting %s\n", psVDisk->m_pzFileName );
	
	int nResult = 0;
	int i = 0;
	char zTmp[10];
	
	LOCK( g_hDiskLock );
	LOCK( psVDisk->m_hLock );
//	printk( " num partitions = %i\n", psVDisk->m_nNumPartitions );
	/* Check if any partitions are still in use */
	for( ; i < MAX_PARTITIONS ; i++ )
	{
		if( psVDisk->m_asPartitions[i].m_nPartitionType != 0 && atomic_read( &psVDisk->m_asPartitions[i].m_nOpenCount ) > 0 ) {
			printk( "VDisk: partition %i is still in use! opencount is %i\n", i-1, atomic_read( &psVDisk->m_asPartitions[i].m_nOpenCount ) );
			nResult = -1;
		}
	}
	if( nResult ) {
		printk( "VDisk: some partitions are still in use! Cannot unmount disk %s while it is still in use.\n", psVDisk->m_pzFileName );
		UNLOCK( psVDisk->m_hLock );
		UNLOCK( g_hDiskLock );
		return( -VDISK_IN_USE );
	}
	for( ; i < MAX_PARTITIONS ; i++ )
	{
		if( psVDisk->m_asPartitions[i].m_nPartitionType != 0 && psVDisk->m_asPartitions[i].m_nDevNodeHandle != 0 ) {
			sprintf( zTmp, "%i", i-1 );
			nResult = delete_device_node( psVDisk->m_asPartitions[i].m_nDevNodeHandle );
//			printk( "VDisk: delete_device_node %i: %s/%s     result = %i\n", psVDisk->m_asPartitions[i].m_nDevNodeHandle, psVDisk->m_pzDevDir, (i == 0 ? "raw" : zTmp ), nResult );
			if( nResult < 0 ) {
				printk( "VDisk: could not delete device node %s/%s !\n", psVDisk->m_pzDevDir, (i == 0 ? "raw" : zTmp ) );
				UNLOCK( psVDisk->m_hLock );
				UNLOCK( g_hDiskLock );
				return( nResult );
			}
		}
	}
	nResult = close( psVDisk->m_nHandle );
	if( nResult < 0 ) {
		printk( "VDisk: could not close file %s! result is %i\n", psVDisk->m_pzFileName, nResult );
		UNLOCK( psVDisk->m_hLock );
		UNLOCK( g_hDiskLock );
		return( nResult );
	}
	printk( "VDisk: successfully unmounted %s.\n", psVDisk->m_pzFileName );
	VirtualDisk_s* psTmp = g_psFirstVDisk;
	if( g_psFirstVDisk == psVDisk ) {
		g_psFirstVDisk = psVDisk->m_psNextVDisk;
	} else {
		while( psTmp != NULL ) {
			if( psTmp->m_psNextVDisk == psVDisk ) {
				psTmp->m_psNextVDisk = psVDisk->m_psNextVDisk;
			}
			psTmp = psTmp->m_psNextVDisk;
		}
	}
	kfree( psVDisk->m_pzDevDir );
	kfree( psVDisk->m_pzFileName );
	UNLOCK( psVDisk->m_hLock );
	delete_semaphore( psVDisk->m_hLock );
	kfree( psVDisk );
	UNLOCK( g_hDiskLock );
	return( 0 );
}

/* Check if the given file is already mounted. Return its VirtualDisk_s if so; NULL otherwise. */
VirtualDisk_s* query_vhd( char* pzFileName )
{
//	printk( "query_vhd( %s )\n", pzFileName );
	LOCK( g_hDiskLock );
	VirtualDisk_s* psTmp = g_psFirstVDisk;
	while( psTmp != NULL && strcmp( pzFileName, psTmp->m_pzFileName ) ) { psTmp = psTmp->m_psNextVDisk; }
	UNLOCK( g_hDiskLock );
	return( psTmp );
}


/* Open the control node - the control node is only used for IOCTL to mount a new virtual disk file */
status_t vdisk_control_open( void* pNode, uint32 nFlags, void** ppCookie )
{
	return( 0 );
}

status_t vdisk_control_close( void* pNode, void* pCookie )
{
	return( 0 );
}

status_t vdisk_control_ioctl( void* pNode, void* pCookie, uint32 nCommand, void* pArgs, bool bFromKernel )
{
	int nResult;
//	printk( "VDisk control: IOCTL nCommand = %i  bFromKernel = %s\n", nCommand, (bFromKernel?"true":"false") );
	
	if( nCommand != VDISK_MOUNT_CMD && nCommand != VDISK_UNMOUNT_CMD && nCommand != VDISK_QUERY_CMD ) {
		printk( "VDisk: invalid IOCTL command %i on control node\n", nCommand );
		return( -EINVAL );
	}
	VDiskIOCTL_s sArgs;
	if( bFromKernel ) {
		memcpy( &sArgs, pArgs, sizeof( VDiskIOCTL_s ) );
	} else {
		memcpy_from_user( &sArgs, pArgs, sizeof( VDiskIOCTL_s ) );
	}
	
//	printk( "  vdisk_control_ioctl: filename is %s, readonly %s, mountpoint %s\n", sArgs.m_zFileName, sArgs.m_bReadOnly? "true":"false", sArgs.m_zMountDir );
	
	VirtualDisk_s* psVDisk = query_vhd( sArgs.m_zFileName );  /* check if the file is already mounted; returns NULL if not */
	if( psVDisk != NULL ) {
		strcpy( sArgs.m_zFileName, psVDisk->m_pzFileName );
		strcpy( sArgs.m_zMountDir, psVDisk->m_pzDevDir );
		sArgs.m_bReadOnly = psVDisk->m_bReadOnly;
		if( bFromKernel ) {
			memcpy( &sArgs, pArgs, sizeof( VDiskIOCTL_s ) );
		} else {
			memcpy_to_user( pArgs, &sArgs, sizeof( VDiskIOCTL_s ) );
		}
	}
	
	switch( nCommand ) {
		case VDISK_QUERY_CMD:
		{
			if( psVDisk == NULL ) {
				return( VDISK_NOT_MOUNTED );
			} else {
				return( VDISK_MOUNTED );
			}
			break;
		}
		case VDISK_MOUNT_CMD:
		{
			if( psVDisk != NULL ) { return( -VDISK_MOUNTED ); }
			else {
				nResult = mount_vhd( sArgs.m_zFileName, sArgs.m_zMountDir, sArgs.m_bReadOnly );
				if( nResult != 0 ) { return( nResult ); }
				psVDisk = query_vhd( sArgs.m_zFileName );
				if( psVDisk != NULL ) {
					strcpy( sArgs.m_zFileName, psVDisk->m_pzFileName );
					strcpy( sArgs.m_zMountDir, psVDisk->m_pzDevDir );
					sArgs.m_bReadOnly = psVDisk->m_bReadOnly;
					if( bFromKernel ) {
						memcpy( &sArgs, pArgs, sizeof( VDiskIOCTL_s ) );
					} else {
						memcpy_to_user( pArgs, &sArgs, sizeof( VDiskIOCTL_s ) );
					}
				} else {
					printk( "VDisk: vdisk_control_ioctl: mount for %s succeeded but query_vhd fails!\n", sArgs.m_zFileName );
				}
				return( nResult );
			}
			break;
		}
		case VDISK_UNMOUNT_CMD:
		{
			if( psVDisk == NULL ) {
				return( -VDISK_NOT_MOUNTED );
			} else {
				return( unmount_vhd( psVDisk ) );
			}
			break;
		}
		default:
			break;
	}
}

status_t vdisk_open( void* pNode, uint32 nFlags, void** pCookie )
{
//	printk( "vdisk_open(_)\n" );
	VirtualPartition_s* psPartition = pNode;
	atomic_inc( &psPartition->m_nOpenCount );
//	printk( "vdisk_open: opencount for partition %i is now %i\n", psPartition->m_nIndex, atomic_read( &psPartition->m_nOpenCount ) );
	return( 0 );
}

status_t vdisk_close( void* pNode, void* pCookie )
{
//	printk( "vdisk_close(_)\n" );
	VirtualPartition_s* psPartition = pNode;
	atomic_dec( &psPartition->m_nOpenCount );
//	printk( "vdisk_close: opencount for partition %i is now %i\n", psPartition->m_nIndex, atomic_read( &psPartition->m_nOpenCount ) );
	return( 0 );
}

status_t vdisk_ioctl( void* pNode, void* pCookie, uint32 nCommand, void* pArgs, bool bFromKernel )
{
//	printk( "vdisk_ioctl( %i, %s )\n", nCommand, (bFromKernel? "true" : "false") );
	int nResult;
	VirtualPartition_s* psPartition = pNode;   /* How to get the correct partition number? */
	VirtualDisk_s* psVDisk = psPartition->m_psDisk;
	LOCK( psVDisk->m_hLock );

	if( nCommand == IOCTL_GET_DEVICE_GEOMETRY ) {
		device_geometry sGeom;
		sGeom.sector_count = psPartition->m_nSize / 512;
		sGeom.cylinder_count = 1;
		sGeom.sectors_per_track = 1;
		sGeom.head_count = 1;
		sGeom.bytes_per_sector = 512;
		sGeom.read_only = psVDisk->m_bReadOnly;
		sGeom.removable = false;
		
		if ( bFromKernel )
		{
			memcpy( pArgs, &sGeom, sizeof( device_geometry ) );
			nResult = 0;
		}
		else
		{
			nResult = memcpy_to_user( pArgs, &sGeom, sizeof( device_geometry ) );
		}
		UNLOCK( psVDisk->m_hLock );
		return( nResult );
	} else if( nCommand == IOCTL_REREAD_PTABLE ) {
		printk( "VDisk: partition table changed IOCTL (not yet implemented)\n" ); /* TODO */
		UNLOCK( psVDisk->m_hLock );
		return( 0 );
	}
	UNLOCK( psVDisk->m_hLock );
	return( -EINVAL );
}

status_t vdisk_read( void* pNode, void* pCookie, off_t nPosition, void* pBuffer, size_t nSize )
{
//	printk( "vdisk_read( _, _, %i, _, %i )\n", nPosition, nSize );
	VirtualPartition_s* psPartition = pNode;
	VirtualDisk_s* psVDisk = psPartition->m_psDisk;
	LOCK( psVDisk->m_hLock );
	size_t nDiskOffset = psPartition->m_nStart + nPosition;

	if( nPosition < 0 || nPosition >= psPartition->m_nSize ) {
		printk( "VDisk: vdisk_read with invalid position %i\n", nPosition );
		UNLOCK( psVDisk->m_hLock );
		return( -EINVAL );
	}
	if( nSize + nPosition > psPartition->m_nSize ) {
		printk( "VDisk: vdisk_read with size %i exceeding partition end; truncating\n", nSize );
		nSize = psPartition->m_nSize - nPosition;
	}
	
	int nResult =  vdisk_pread( psVDisk, pBuffer, nSize, nDiskOffset );
	UNLOCK( psVDisk->m_hLock );
	return( nResult );
}

status_t vdisk_write( void* pNode, void* pCookie, off_t nPosition, const void* pBuffer, size_t nSize )
{
//	printk( "vdisk_write( _, _, %li, _, %li )\n", nPosition, nSize );
	VirtualPartition_s* psPartition = pNode;
	VirtualDisk_s* psVDisk = psPartition->m_psDisk;
	LOCK( psVDisk->m_hLock );
	size_t nDiskOffset = psPartition->m_nStart + nPosition;

	if( psVDisk->m_bReadOnly ) { UNLOCK( psVDisk->m_hLock ); return( -EROFS ); }
	if( nPosition < 0 || nPosition >= psPartition->m_nSize ) {
		printk( "VDisk: vdisk_write with invalid position %i\n", nPosition );
		UNLOCK( psVDisk->m_hLock );
		return( -EINVAL );
	}
	if( nSize + nPosition > psPartition->m_nSize ) {
		printk( "VDisk: vdisk_write with size %i exceeding partition end; truncating\n", nSize );
		nSize = psPartition->m_nSize - nPosition;
	}

	int nResult = vdisk_pwrite( psVDisk, pBuffer, nSize, nDiskOffset );
	UNLOCK( psVDisk->m_hLock );
	return( nResult );
}

status_t vdisk_readv( void* pNode, void* pCookie, off_t nPosition, const struct iovec* psVector, size_t nCount )
{
//	printk( "vdisk_readv(...)\n" );
	VirtualPartition_s* psPartition = pNode;
	VirtualDisk_s* psVDisk = psPartition->m_psDisk;
	LOCK( psVDisk->m_hLock );
	size_t nDiskOffset = psPartition->m_nStart + nPosition;
	
	size_t i = 0;
	size_t nTotalSize = 0;
	size_t nBytesRead = 0;
	size_t nResult;
	if( nCount < 0 ) return( -EINVAL );
	for( ; i < nCount; i++ ) {
		if( psVector[i].iov_len < 0 ) {
			printk( "VDisk: vdisk_readv with negative length\n" );
			UNLOCK( psVDisk->m_hLock );
			return( -EINVAL );
		}
		nTotalSize += psVector[i].iov_len;
	}
	if( nPosition < 0 || nPosition >= psPartition->m_nSize ) {
		printk( "VDisk: vdisk_readv with invalid position %i\n", nPosition );
		UNLOCK( psVDisk->m_hLock );
		return( -EINVAL );
	}
	if( nTotalSize + nPosition > psPartition->m_nSize ) {
		printk( "VDisk: vdisk_readv with total size %i exceeding partition end; truncating\n", nTotalSize );
		nTotalSize = psPartition->m_nSize - nPosition;
	}

	for( i = 0 ; i < nCount; i++ ) {
		if( nBytesRead + psVector[i].iov_len > nTotalSize ) {
			nResult = vdisk_pread( psVDisk, psVector[i].iov_base, (nTotalSize - nBytesRead), nDiskOffset );
			if( nResult < 0 ) {
				UNLOCK( psVDisk->m_hLock );
				return( nResult );
			}
			nBytesRead += nResult;
			UNLOCK( psVDisk->m_hLock );
			return( nBytesRead );
		}
		nResult = vdisk_pread( psVDisk, psVector[i].iov_base, psVector[i].iov_len, nDiskOffset );
		if( nResult < 0 )
		{
			UNLOCK( psVDisk->m_hLock );
			return( nResult );
		}
		nDiskOffset += nResult;
		nBytesRead += nResult;
	}
	UNLOCK( psVDisk->m_hLock );
	return( nBytesRead );
}

status_t vdisk_writev( void* pNode, void* pCookie, off_t nPosition, const struct iovec* psVector, size_t nCount )
{
//	printk( "vdisk_writev(...)\n" );
	VirtualPartition_s* psPartition = pNode;
	VirtualDisk_s* psVDisk = psPartition->m_psDisk;
	LOCK( psVDisk->m_hLock );
	size_t nDiskOffset = psPartition->m_nStart + nPosition;

	size_t i = 0;
	size_t nTotalSize = 0;
	size_t nBytesWritten = 0;
	size_t nResult;
	if( psVDisk->m_bReadOnly ) { UNLOCK( psVDisk->m_hLock ); return( -EROFS ); }
	if( nCount < 0 ) { UNLOCK( psVDisk->m_hLock ); return( -EINVAL ); }
	for( ; i < nCount; i++ ) {
		if( psVector[i].iov_len < 0 ) {
			printk( "VDisk: vdisk_writev with negative length\n" );
			UNLOCK( psVDisk->m_hLock );
			return( -EINVAL );
		}
		nTotalSize += psVector[i].iov_len;
	}
	if( nPosition < 0 || nPosition >= psPartition->m_nSize ) {
		printk( "VDisk: vdisk_writev with invalid position %i\n", nPosition );
		UNLOCK( psVDisk->m_hLock );
		return( -EINVAL );
	}
	if( nTotalSize + nPosition > psPartition->m_nSize ) {
		printk( "VDisk: vdisk_writev with total size %i exceeding partition end; truncating\n", nTotalSize );
		nTotalSize = psPartition->m_nSize - nPosition;
	}

	for( i = 0 ; i < nCount; i++ ) {
		if( nBytesWritten + psVector[i].iov_len > nTotalSize ) {
			nResult = vdisk_pwrite( psVDisk, psVector[i].iov_base, (nTotalSize - nBytesWritten), nDiskOffset );
			if( nResult < 0 ) {
				return( nResult );
				UNLOCK( psVDisk->m_hLock );
			}
			nBytesWritten += nResult;
			UNLOCK( psVDisk->m_hLock );
			return( nBytesWritten );
		}
		nResult = vdisk_pwrite( psVDisk, psVector[i].iov_base, psVector[i].iov_len, nDiskOffset );
		if( nResult < 0 )
		{
			UNLOCK( psVDisk->m_hLock );
			return( nResult );
		}
		nDiskOffset += nResult;
		nBytesWritten += nResult;
	}
	UNLOCK( psVDisk->m_hLock );
	return( nBytesWritten );
}

status_t device_init( int nDeviceID )
{
//	printk( "VDisk: device_init( %i )\n", nDeviceID );
	printk( "VDisk virtual disk driver v" VDISK_VERSION " loaded\n" );
	
	g_nDeviceID = nDeviceID;
	
	g_hDiskLock = create_semaphore( "vdisk_lisk_lock", 1, 0 );
	if( g_hDiskLock < 0 ) {
		printk( "VDisk: could not create disk list lock!\n" );
		return( g_hDiskLock );
	}
	int nResult = create_device_node( nDeviceID, -1, "disk/virtual/control", &g_sVControlOps, NULL );
	if( nResult <= 0 ) {
		printk( "VDisk: unable to create control device! result is %i\n", nResult );
		return( nResult );
	}
	g_nControlDevNodeHandle = nResult;
	return( 0 );
}

status_t device_uninit( int nDeviceID )
{
//	printk( "VDisk: device_uninit( %i )\n", nDeviceID );
	int nResult;
	while( g_psFirstVDisk != NULL ) {
		nResult = unmount_vhd( g_psFirstVDisk );
		if( nResult ) {
			printk( "VDisk: failed to unmount disk %s in device_uninit!\n", g_psFirstVDisk->m_pzFileName );
			return( nResult );
		}
	}
	
	delete_device_node( g_nControlDevNodeHandle );
	delete_semaphore( g_hDiskLock );
	printk( "VDisk driver unloaded.\n" );
	return( 0 );
}








