#ifndef __VIRTUALDISK_H__
#define __VIRTUALDISK_H__

/* The maximum number of partitions we will create device nodes for */
#define MAX_PARTITIONS 16

/* Disk geometry struct */
typedef struct {
	uint16  m_nCylinders;
	uint8   m_nHeads;
	uint8   m_nSectorsPerTrack;
} VHDDiskGeom_s;

/* Microsoft VHD footer */
typedef struct {
	char     m_zCookie[8];
	uint32   m_nFeatures;
	uint32   m_nFileFormatVersion;
	uint64   m_nDataOffset;
	uint32   m_nTimeStamp;
	char     m_zCreatorApp[4];
	uint32   m_nCreatorAppVersion;
	char     m_zCreatorOSHost[4];
	uint64   m_nOriginalSize;
	uint64   m_nCurrentSize;
	VHDDiskGeom_s m_sDiskGeometry;
	uint32   m_nDiskType;
	uint32   m_nChecksum;
	uint8    m_aUUID[16];
	uint8    m_nSavedState;
	char     m_zReserved[427];
} VHDFooter_s;

typedef struct VirtualDisk_struct VirtualDisk_s;

typedef struct VirtualPartition_struct {
	VirtualDisk_s*  m_psDisk;	/* Pointer to the parent disk struct */
	int		 m_nDevNodeHandle;
	int		 m_nIndex;		    /* Index of the partition in disk's partition array */
	size_t   m_nStart;           /* Start of the disk, in bytes from the start of the virtual disk */
	size_t   m_nSize;			/* Size of the partition in bytes */
	uint8	 m_nStatus;			/* Fields from the partition table */
	uint8	 m_nPartitionType;
	atomic_t m_nOpenCount;
} VirtualPartition_s;

struct VirtualDisk_struct {
	int    m_nHandle;
	char*  m_pzFileName;
	char*  m_pzDevDir;
	bool   m_bReadOnly;
	size_t m_nDiskSize;        /* disk size, in bytes */
	int	   m_nNumPartitions;   /* The number of valid partitions on the disk, not including partition raw (the whole disk) */
	VirtualPartition_s m_asPartitions[MAX_PARTITIONS + 1];
	sem_id m_hLock;
	VHDFooter_s m_sFooter;
	VirtualDisk_s*  m_psNextVDisk;
};

//#pragma pack(push)
#pragma pack(1)			/* These structures must be byte-aligned */

/* MBR Partiton table entry - also used by extended partitions */
typedef struct {
	uint8  m_nStatus;        /* Bootable flag */
	uint8  m_nStartHead;     /* Head of the first sector of the partition */
	uint16 m_nStartCylSec;   /* Cylinder & sector of first sector; encoded wierdly */
	uint8  m_nPartitionType; /* Type ID of partiton */
	uint8  m_nEndHead;       /* Head of the last sector of the partition */
	uint16 m_nEndCylSec;     /* Cylinder & sector of end of partition */
	uint32 m_nLBAStart;      /* First sector of the partition */
	uint32 m_nLBALength;     /* number of sectors in the partition */
} MBRPartitionEntry_s;

/* Master Boot Record / Partition table */
typedef struct {
	uint8  m_aBootCode[446];  /* Boot asm code */
	MBRPartitionEntry_s  m_aPartitionEntry[4];   /* Partition table */
	uint16 m_nMBRId;          /* 0x55aa signature */
} MBR_s;

#pragma pack()

/* Utility functions */
int read_vhd_partition_table( VirtualDisk_s* psVDisk, size_t nMBRLocation, size_t nDiskSize, int nPartitionsReadSoFar );
uint32 compute_checksum( VHDFooter_s* psFooter );
int pread( int nHandle, void* pBuffer, size_t nLength, size_t nOffset );  /* Regular pread function - absent from kernel API */
int vdisk_pread( VirtualDisk_s* psVDisk, void* pBuffer, size_t nLength, size_t nOffset );  /* Read data from virtual disk */
void byteswap( void* pWhere, int nSize );


/* Operations for the control node, /dev/disk/virtual/control */
status_t vdisk_control_open( void* pNode, uint32 nFlags, void** ppCookie );
status_t vdisk_control_close( void* pNode, void* pCookie );
status_t vdisk_control_ioctl( void* pNode, void* pCookie, uint32 nCommand, void* pArgs, bool bFromKernel );

/* Operations for mounted virtual disks */
status_t vdisk_open( void* pNode, uint32 nFlags, void** ppCookie );
status_t vdisk_close( void* pNode, void* pCookie );
status_t vdisk_ioctl( void* pNode, void* pCookie, uint32 nCommand, void* pArgs, bool bFromKernel );
status_t vdisk_read( void* pNode, void* pCookie, off_t nPosition, void* pBuffer, size_t nSize );
status_t vdisk_write( void* pNode, void* pCookie, off_t nPosition, const void* pBuffer, size_t nSize );
status_t vdisk_readv( void* pNode, void* pCookie, off_t nPosition, const struct iovec* psVector, size_t nCount );
status_t vdisk_writev( void* pNode, void* pCookie, off_t nPosition, const struct iovec* psVector, size_t nCount );

/* Driver init & uninit */
status_t device_init( int nDeviceID );
status_t device_uninit( int nDeviceID );



#endif /* __VIRTUALDISK_H__ */



