/* Util to mount and unmount VHD files using the VirtualDisk driver for Syllable OS.
   Performs an IOCTL on /dev/disk/virtual/control 										*/

/* Written by Anthony Morphett  awmorp@gmail.com   July 2007*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <sys/ioctl.h>

#include "../driver/protocol.h"

void usage()
{
	printf( "mountvhd: a utility to mount and unmount VirtualPC VHD files.\n" );
	printf( "Version " VDISK_VERSION "\n\n" );
	printf( "Usage:\n" );
	printf( "To mount the virtual hard drive VHDFILE:\n" );
	printf( "    mountvhd -m [ -r | -w ] [-d MOUNT-DIR ] VHDFILE\n" );
	printf( "       Options:\n" );
	printf( "        -r/-w: mount read-only/read-write (default is read-write)\n" );
	printf( "        -d MOUNT-DIR: mount the virtual drive at /dev/disk/virtual/MOUNT-DIR (default is same as VHDFILE).\n" );
	printf( "\n" );
	printf( "To unmount the virtual hard drive VHDFILE:\n" );
	printf( "    mountvhd -u VHDFILE\n" );
	printf( "\n" );
	printf( "To check if the virtual hard drive VHDFILE is already mounted (this is the default):\n" );
	printf( "    mountvhd -q VHDFILE\n" );
}


int main( int nArgc, char* pzArgv[] )
{
	VDiskIOCTL_s sArg;
	int nAction = VDISK_QUERY_CMD;
	int nResult;
	int nHandle;
	
	sArg.m_zFileName[0] = 0;
	sArg.m_zMountDir[0] = 0;
	sArg.m_bReadOnly = false;
	
	int c;
	while( ( c = getopt( nArgc, pzArgv, "muqrwhd:" ) ) != -1 )
	{
		switch ( c )
		{
		case 0:
			break;
		case ':':
		case '?':
			usage();
			return( 1 );
			break;
		case 'm':
			nAction = VDISK_MOUNT_CMD;
			break;
		case 'u':
			nAction = VDISK_UNMOUNT_CMD;
			break;
		case 'q':
			nAction = VDISK_QUERY_CMD;
			break;
		case 'r':
			sArg.m_bReadOnly = true;
			break;
		case 'w':
			sArg.m_bReadOnly = false;
			break;
		case 'd':
			strncpy( sArg.m_zMountDir, optarg, 255 );
			sArg.m_zMountDir[255] = 0;
			break;
		default:
			usage();
			return( 1 );
			break;
		}
	}
	
	if( optind >= nArgc )
	{
		printf( "Error: a filename is required!\n" );
		usage();
		return( 1 );
	}

	strncpy( sArg.m_zFileName, pzArgv[optind], 255 );
	sArg.m_zFileName[255] = 0;

	switch( nAction ) {
		case VDISK_MOUNT_CMD:
		{
			printf( "Mounting virtual drive %s%s", sArg.m_zFileName, (sArg.m_bReadOnly ? " read-only" : "") );
			if( sArg.m_zMountDir[0] ) { printf( " to /dev/disk/virtual/%s", sArg.m_zMountDir ); }
			printf( "...\n" );
			nHandle = open( "/dev/disk/virtual/control", O_RDONLY );
			if( nHandle < 0 ) {
				printf( "Error: could not open control node /dev/disk/virtual/control ! There is possibly a problem with the driver.\n" );
				printf( "Check the kernel log for any error messages.\n" );
				printf( "The error message is %s\n", strerror( nHandle ) );
				return( 2 );
			}
			nResult = ioctl( nHandle, nAction, &sArg );
			if( nResult == -VDISK_MOUNTED ) {
				printf( "The virtual disk %s is already mounted at /dev/%s.\n", sArg.m_zFileName, sArg.m_zMountDir );
				close( nHandle );
				return( 1 );
			} else if( nResult < 0 ) {
				printf( "Error: IOCTL to control node /dev/disk/virtual/control failed ! There is possibly a problem with the driver.\n" );
				printf( "Check the kernel log for any error messages.\n" );
				printf( "The error message is %s\n", strerror( nResult ) );
				close( nHandle );
				return( 2 );
			} else if( nResult == 0 ) {
				printf( "Virtual disk %s was successfully mounted%s at /dev/%s.\n", sArg.m_zFileName, (sArg.m_bReadOnly ? " read-only" : ""), sArg.m_zMountDir );
				close( nHandle );
				return( 0 );
			} else {
				printf( "Error: unknown return code %i from IOCTL!\n", nResult );
				close( nHandle );
				return( 2 );
			}
			break;
		}	
		case VDISK_UNMOUNT_CMD:
		{
			printf( "Unmounting virtual drive %s...\n", sArg.m_zFileName );
			nHandle = open( "/dev/disk/virtual/control", O_RDONLY );
			if( nHandle < 0 ) {
				printf( "Error: could not open control node /dev/disk/virtual/control ! There is possibly a problem with the driver.\n" );
				printf( "Check the kernel log for any error messages.\n" );
				printf( "The error message is %s\n", strerror( nHandle ) );
				return( 2 );
			}
			nResult = ioctl( nHandle, nAction, &sArg );
			if( nResult == -VDISK_NOT_MOUNTED ) {
				printf( "Error: the virtual disk %s is not mounted! Run 'mountvhd -m %s' to mount it.\n", sArg.m_zFileName, sArg.m_zFileName );
				close( nHandle );
				return( 1 );
			} else if( nResult == -VDISK_IN_USE ) {
				printf( "Error: the virtual disk %s is still in use, so it could not be unmounted.\n", sArg.m_zFileName );
				return( 1 );
			} else if( nResult < 0 ) {
				printf( "Error: IOCTL to control node /dev/disk/virtual/control failed ! There is possibly a problem with the driver.\n" );
				printf( "Check the kernel log for any error messages.\n" );
				printf( "The error message is %s\n", strerror( nResult ) );
				close( nHandle );
				return( 2 );
			} else if( nResult == 0 ) {
				printf( "Virtual disk %s was successfully unmounted.\n", sArg.m_zFileName );
				close( nHandle );
				return( 0 );
			} else {
				printf( "Error: unknown return code %i from IOCTL!\n", nResult );
				close( nHandle );
				return( 2 );
			}
			break;
		}
		case VDISK_QUERY_CMD:
		{
			printf( "Querying virtual drive %s...\n", sArg.m_zFileName );
			nHandle = open( "/dev/disk/virtual/control", O_RDONLY );
			if( nHandle < 0 ) {
				printf( "Error: could not open control node /dev/disk/virtual/control ! There is possibly a problem with the driver.\n" );
				printf( "Check the kernel log for any error messages.\n" );
				printf( "The error message is %s\n", strerror( nHandle ) );
				return( 2 );
			}
			nResult = ioctl( nHandle, nAction, &sArg );
			if( nResult < 0 ) {
				printf( "Error: IOCTL to control node /dev/disk/virtual/control failed ! There is possibly a problem with the driver.\n" );
				printf( "Check the kernel log for any error messages.\n" );
				printf( "The error message is %s\n", strerror( nResult ) );
				close( nHandle );
				return( 2 );
			} else if( nResult == VDISK_MOUNTED ) {
				printf( "Virtual disk %s is mounted %s at %s.\n", sArg.m_zFileName, (sArg.m_bReadOnly ? "read-only" : "read-write"), sArg.m_zMountDir );
				close( nHandle );
				return( 0 );
			} else if( nResult == VDISK_NOT_MOUNTED ) {
				printf( "Virtual disk %s is not mounted.\n", sArg.m_zFileName );
				close( nHandle );
				return( 1 );
			} else {
				printf( "Error: unknown return code %i from IOCTL!\n", nResult );
				close( nHandle );
				return( 2 );
			}
		}
	}
}




